// sendKeystrokes.cpp : This file contains the 'main' function. Program execution begins and ends there.
//
// Expected MATLAB IDE keyboard shortcuts:
//  Ctrl+Shift+0 = Activate code-window (default, cannot be changed in Matlab R2017a)
//  Ctrl+o = Open file
//  Ctrl+g = Go to line
//  F12 = Toggle breakpoint
//


#include <Windows.h>
#include <stdio.h>
#include <assert.h>


HWND findWindow(const wchar_t* strTitle, bool bWaitForIt, int nTimeout);
void sendKey(SHORT key, bool isVk = false);
void sendText(const wchar_t* strText);
void parseArgs(int argc, wchar_t* argv[]);


#define _strMainWndTitle L"MATLAB R2017a"
#define _strFileToGo L"C:\\path\\file.m"
#define _strLineToGo L"0"
#define keyDelayMs 10
#define guiUpdateDelayMs 400


#define MODIF_CTL 0x1000
#define MODIF_SFT 0x2000
#define MODIF_ALT 0x4000
#define MODIF_ALG 0x8000


wchar_t strMainWndTitle[64];
wchar_t strFileToGo[1024];
wchar_t strLineToGo[16];
bool bAddBreakpoint;
bool bOnlyBreakpoint;
bool bOnlyGo;

// sendKeystrokes.exe "C:\path\file.m" 83 BREAK "MATLAB R2017a"
// sendKeystrokes.exe "C:\path\file.m" 83 BREAK
// sendKeystrokes.exe "C:\path\file.m" 83
// sendKeystrokes.exe "C:\path\file.m"
// sendKeystrokes.exe GO
// sendKeystrokes.exe BREAK
int wmain(int argc, wchar_t* argv[]) {

	HWND hwndConsole = GetConsoleWindow();
	if (!hwndConsole) { exit(-2); return 0; }
	parseArgs(argc, argv);

	HWND h = findWindow(strMainWndTitle, true, 3);
	assert(h);
	ShowWindow(hwndConsole, SW_HIDE);

	keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0); // Control release
	keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0); // Shift release
	keybd_event(VK_RMENU, 0, KEYEVENTF_KEYUP, 0); // AltGr release
	keybd_event(VK_LMENU, 0, KEYEVENTF_KEYUP, 0); // normal Alt release
	Sleep(keyDelayMs);

	//check if already has focus
	if (h != GetForegroundWindow()) {
		//check if window is minimized
		if (IsIconic(h)) {
			ShowWindow(h, SW_RESTORE);
			Sleep(guiUpdateDelayMs);
			sendKey(MODIF_ALT);
		}
		SetForegroundWindow(h);
	}

	sendKey(MODIF_CTL | MODIF_SFT | '0');
	Sleep(keyDelayMs);

	if (bOnlyBreakpoint) {
		Sleep(guiUpdateDelayMs);
		sendKey(VK_F12, true);
		return 0;
	} else if (bOnlyGo) {
		Sleep(guiUpdateDelayMs);
		sendKey(VK_F5, true);
		return 0;
	}
	sendKey(MODIF_CTL | 'o');
	Sleep(guiUpdateDelayMs);
	sendText(strFileToGo);
	sendKey(VK_RETURN, true);
	int line = _wtoi(strLineToGo);
	if (line > 0 && line < 99999) {
		swprintf_s(strLineToGo, sizeof(strLineToGo) / sizeof(strLineToGo[0]), L"%d", line);
		Sleep(2*guiUpdateDelayMs);
		sendKey(MODIF_CTL | 'g');
		sendText(strLineToGo);
		sendKey(VK_RETURN, true);
	}
	if (bAddBreakpoint) {
		Sleep(guiUpdateDelayMs);
		sendKey(VK_F12, true);
	}

#ifdef _DEBUG
	ShowWindow(hwndConsole, SW_SHOW);
	wprintf(L"\nDone!\n");
	system("pause");
#endif // _DEBUG

	return 0;
}


HWND findWindow(const wchar_t* strTitle, bool bWaitForIt, int nTimeout) {
	HWND hwnd = FindWindow(NULL, strTitle);
	wchar_t buff[256];
	while (!hwnd && nTimeout > 0) {
		swprintf_s(buff, sizeof(buff)/sizeof(buff[0]), L"Couldn't find \"%s\" window... (%d). Please try to manually open it!", strTitle, nTimeout);
		MessageBox(NULL, buff, L"Find window", MB_OK | MB_ICONEXCLAMATION);
		nTimeout--;
		hwnd = FindWindow(NULL, strTitle);
	}
	if (!hwnd) {
		swprintf_s(buff, sizeof(buff) / sizeof(buff[0]), L"Couldn't find \"%s\" window... Exiting!", strTitle);
		MessageBox(NULL, buff, L"Find window", MB_OK | MB_ICONERROR);
		// system("pause");
		exit(-1); return 0;
	}
	return hwnd;
}


void sendKey(SHORT key, bool isVk) {

	SHORT vk;
	if (isVk) {
		vk = (BYTE)key;
	} else {
		vk = VkKeyScan((BYTE)key);
	}
	if (vk == 0xFFFF) return;
	bool bSendSft = (vk & 0x100); // check upper byte for shift flag
	bool bSendCtl = ((vk & 0x600) == 0x200); // check for ctrl flag
	bool bSendAlt = ((vk & 0x600) == 0x400); // check for normal alt flag
	bool bSendAlG = ((vk & 0x600) == 0x600); // check for AltGr flag
	bSendCtl = bSendCtl ^ (bool)(key & MODIF_CTL);
	bSendSft = bSendSft ^ (bool)(key & MODIF_SFT);
	bSendAlt = bSendAlt ^ (bool)(key & MODIF_ALT);
	bSendAlG = bSendAlG ^ (bool)(key & MODIF_ALG);
	BYTE nVks = (BYTE)vk;

	wprintf(L"sendKey: ");
	if (bSendCtl) {
		keybd_event(VK_CONTROL, 0, 0, 0); // Control press
		wprintf(L"Ctl+");
	}
	if (bSendSft) {
		keybd_event(VK_SHIFT, 0, 0, 0); // Shift press
		wprintf(L"Sft+");
	}
	if (bSendAlG) {
		keybd_event(VK_RMENU, 0, 0, 0); // AltGr press
		wprintf(L"AlG+");
	}
	if (bSendAlt) {
		keybd_event(VK_LMENU, 0, 0, 0); // normal Alt press
		wprintf(L"Alt+");
	}

	if (nVks) {
		keybd_event(nVks, 0, 0, 0); // key press
		switch (key)
		{
		case VK_RETURN:
			wprintf(L"Enter\n");
			break;
		case VK_F12:
			wprintf(L"F12\n");
			break;
		case VK_F5:
			wprintf(L"F5\n");
			break;
		default:
			wprintf(L"%c\n", (BYTE)key);
			break;
		}
		Sleep(2*keyDelayMs);
		keybd_event(nVks, 0, KEYEVENTF_KEYUP, 0); // key release
	}

	if (bSendCtl) keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0); // Control release
	if (bSendSft) keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0); // Shift release
	if (bSendAlG) keybd_event(VK_RMENU, 0, KEYEVENTF_KEYUP, 0); // AltGr release
	if (bSendAlt) keybd_event(VK_LMENU, 0, KEYEVENTF_KEYUP, 0); // normal Alt release
}


void sendText(const wchar_t* strText) {
	//HKL hkl = GetKeyboardLayout(NULL);
	int L = (int)wcslen(strText);
	if (L > 1024) return;
	for (int i = 0; i < L; ++i) {
		//SHORT vk = VkKeyScanEx(strText[i], hkl);
		SHORT vk = VkKeyScan(strText[i]);
		if (vk == 0xFFFF) break;
		bool bSendSft = (vk & 0x100); // check upper byte for shift flag
		bool bSendCtl = ((vk & 0x600) == 0x200); // check for ctrl flag
		bool bSendAlt = ((vk & 0x600) == 0x400); // check for normal alt flag
		bool bSendAlG = ((vk & 0x600) == 0x600); // check for AltGr flag
		BYTE code = (BYTE)vk;
		if (bSendCtl) keybd_event(VK_CONTROL, 0, 0, 0); // Ctrl press
		if (bSendSft) keybd_event(VK_SHIFT,   0, 0, 0); // Shift press
		if (bSendAlG) keybd_event(VK_RMENU,   0, 0, 0); // AltGr press
		if (bSendAlt) keybd_event(VK_LMENU,   0, 0, 0); // normal Alt press
		keybd_event(code, 0, 0, 0); // key press
		Sleep(keyDelayMs);
		keybd_event(code, 0, KEYEVENTF_KEYUP, 0); // key release
		if (bSendCtl) keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0); // Ctrl release
		if (bSendSft) keybd_event(VK_SHIFT,   0, KEYEVENTF_KEYUP, 0); // Shift release
		if (bSendAlG) keybd_event(VK_RMENU,   0, KEYEVENTF_KEYUP, 0); // AltGr release
		if (bSendAlt) keybd_event(VK_LMENU,   0, KEYEVENTF_KEYUP, 0); // normal Alt release
	}
	wprintf(L"sentText: \"%s\"\n", strText);
}


void parseArgs(int argc, wchar_t* argv[]) {
	if (argc > 5 || argc < 2) {
		wprintf(L"Detected %d parameters!\n"
			"Usage:\n"
			" \"%s\" \"%s\" %s BREAK \"%s\"\n"
			" \"%s\" \"%s\" %s BREAK\n"
			" \"%s\" \"%s\" %s\n"
			" \"%s\" \"%s\"\n"
			" \"%s\" GO\n"
			" \"%s\" BREAK\n"
			"Needed MATLAB IDE keyboard shortcuts:\n"
			"  Ctrl+Shift+0 = Activate code-window\n"
			"  Ctrl+o = Open file\n"
			"  Ctrl+g = Go to line\n"
			"  F12 = Toggle breakpoint\n",
			argc,
			argv[0], _strFileToGo, _strLineToGo, _strMainWndTitle,
			argv[0], _strFileToGo, _strLineToGo, 
			argv[0], _strFileToGo, _strLineToGo,
			argv[0], _strFileToGo,
			argv[0],
			argv[0]
			);
		MessageBox(NULL, L"Wrong command line parameters", L"Check usage", MB_OK | MB_ICONEXCLAMATION);
		//system("pause");
		exit(-1); 
		return;
	}
	if (argc == 5) {
		swprintf_s(strMainWndTitle, sizeof(strMainWndTitle) / sizeof(strMainWndTitle[0]), L"%s", argv[4]);
	}
	else {
		swprintf_s(strMainWndTitle, sizeof(strMainWndTitle) / sizeof(strMainWndTitle[0]), L"%s", _strMainWndTitle);
	}
	if (argc >= 4 && _wcsicmp(argv[3], L"BREAK")==0) {
		bAddBreakpoint = true;
	}
	else {
		bAddBreakpoint = false;
	}
	if (argc >= 3) {
		swprintf_s(strLineToGo, sizeof(strLineToGo) / sizeof(strLineToGo[0]), L"%s", argv[2]);
	}
	else {
		swprintf_s(strLineToGo, sizeof(strLineToGo) / sizeof(strLineToGo[0]), L"%s", _strLineToGo);
	}
	if (_wcsicmp(argv[1], L"BREAK") == 0) {
		bOnlyBreakpoint = TRUE;
	} else if (_wcsicmp(argv[1], L"GO") == 0) {
		bOnlyGo = TRUE;
	} else {
		swprintf_s(strFileToGo, sizeof(strFileToGo) / sizeof(strFileToGo[0]), L"%s", argv[1]);
	}
	wprintf(L"Running: \"%s\" \"%s\" \"%s\" \"%s\" bAddBreakpoint=%d bOnlyBreakpoint=%d bOnlyGo=%d\n\n", 
		argv[0], strFileToGo, strLineToGo, strMainWndTitle, bAddBreakpoint, bOnlyBreakpoint, bOnlyGo);
}

